
#property description "Blahtech Supply Demand Expert Advisor - Advanced"
#property version     "2.5"
#property strict

#include "BlahtechSupDemLevel.mqh"

#ifdef __MQL4__
input string            IndicatorPath = "Market\\Blahtech Supply Demand";          // Indicator Path
#else
input string            IndicatorPath = "Market\\Blahtech Supply Demand MT5.ex5";  // Indicator Path
#endif
input SdEaModeEnum      ExpertMode = SD_EA_EXPERT_NO_DRAW;                         // Expert Mode
input int               MinBarsLevel = 3;                                          // Min Bars For Level
input int               MaxBars = 20000;                                           // Max Bars
input int               ForceRecalcBars = 10000;                                   // Force Recalc Bars
input string            LoadParameterFile = "SD_EA_Params.set";                    // Parameter File/Values
//input string            LoadParameterFile =  "values:DetectionMethod2=0;Hud2=2";   // Parameter File/Values
//input string            LoadParameterFile =  "values:Timeframe2=30;Timeframe3=60";   // Parameter File/Values
input SdEaRetouchMode   UpdateRetouches = SD_EA_RETOUCH_FIRST;                     // Update Retouches Mode
input LogLevelEnum      LogLevel = LogError;                                       // Log Level   
input SdEaValidateEnum  ValidateIncremental = SD_EA_VAL_NONE;                      // Validate Incremental

bool              ValidateActive;
int               ExtIndicatorHandle;
SupDemLevel*      ExtSupLevels1[];
SupDemLevel*      ExtDemLevels1[];
SupDemLevel*      ExtSupLevels2[];
SupDemLevel*      ExtDemLevels2[];
SupDemLevel*      ExtSupLevels3[];
SupDemLevel*      ExtDemLevels3[];
SupDemLevelIdx*   ExtLevelIndex1;
SupDemLevelIdx*   ExtLevelIndex2;
SupDemLevelIdx*   ExtLevelIndex3;


double iCustomSd(SdBufferIndexEnum index, int shift = 0)
{
#ifdef __MQL4__
   return 
      iCustom(
         NULL, PERIOD_CURRENT, IndicatorPath,
         ExpertMode, MinBarsLevel, MaxBars, ForceRecalcBars, LoadParameterFile,
         index, shift);
#else
   double arr[1];
   int    retIdx = CopyBuffer(ExtIndicatorHandle, index, shift, 1, arr);
   return retIdx > 0 ? arr[0] : EMPTY_VALUE;
#endif
}


int OnInit()
{
   if (LogLevel >= LogInfo) { Print("OnInit - Start"); }
   
   string            tfMap1 = "M1=M5,M5=M30,M15=H1,M30=H4,H1=H4,H4=D1,D1=W1,W1=MN1,MN1=Q1";
   string            tfMap2 = "M1=M5,M5=M30,M15=H1,M30=H4,H1=H4,H4=D1,D1=W1,W1=MN1,MN1=Q1";
   string            tfMap3 = "M1=M30,M5=H1,M15=H4,M30=D1,H1=D1,H4=W1,D1=MN1,W1=Q1,MN1=Y1";
   MAPPED_TIMEFRAMES mapTimeframe1 = MAP_PERIOD_CURRENT;
   MAPPED_TIMEFRAMES mapTimeframe2 = MAP_PERIOD_MAPPED;
   MAPPED_TIMEFRAMES mapTimeframe3 = MAP_PERIOD_MAPPED;
   
   ExtLevelIndex1 = new SupDemLevelIdx(1, MappedTimeframeToTimeframe(tfMap1, mapTimeframe1), SD_SUPBARNUM,  SD_SUPLOW,  SD_DEMBARNUM,  SD_DEMHIGH);
   ExtLevelIndex2 = new SupDemLevelIdx(2, MappedTimeframeToTimeframe(tfMap2, mapTimeframe2), SD_SUPBARNUM2, SD_SUPLOW2, SD_DEMBARNUM2, SD_DEMHIGH2);
   ExtLevelIndex3 = new SupDemLevelIdx(3, MappedTimeframeToTimeframe(tfMap3, mapTimeframe3), SD_SUPBARNUM3, SD_SUPLOW3, SD_DEMBARNUM3, SD_DEMHIGH3);
   ValidateActive = ValidateIncremental != SD_EA_VAL_NONE;
      
#ifdef __MQL4__
   int  ret = INIT_SUCCEEDED;
#else
   ExtIndicatorHandle =
      iCustom(
         NULL, PERIOD_CURRENT, IndicatorPath,
         ExpertMode, MinBarsLevel, MaxBars, ForceRecalcBars, LoadParameterFile);
         
   int  ret = ExtIndicatorHandle > 0 ? INIT_SUCCEEDED : INIT_FAILED;
#endif
   if (LogLevel >= LogInfo) { Print("OnInit - IndicatorPath=", IndicatorPath); }
   return ret;
}


void OnDeinit(const int reason)
{
   ArrayDelete(ExtSupLevels1);
   ArrayDelete(ExtDemLevels1);
   ArrayDelete(ExtSupLevels2);
   ArrayDelete(ExtDemLevels2);
   ArrayDelete(ExtSupLevels3);
   ArrayDelete(ExtDemLevels3);
   if (CheckPointer(ExtLevelIndex1) == POINTER_DYNAMIC) delete ExtLevelIndex1;
   if (CheckPointer(ExtLevelIndex2) == POINTER_DYNAMIC) delete ExtLevelIndex2;
   if (CheckPointer(ExtLevelIndex3) == POINTER_DYNAMIC) delete ExtLevelIndex3;
#ifndef __MQL4__
   if (ExtIndicatorHandle > 0)
   {
      IndicatorRelease(ExtIndicatorHandle);
   }
#endif
}


void OnTick()
{
   static int      stTickCount;
   static datetime stPrevTime0;
   datetime        time0 = iTime(NULL, 0, 0);
   int             sdChanged = (int)iCustomSd(SD_CHANGED);  // Only do work when values have changed

   stTickCount++;
#ifndef __MQL4__
   if (stTickCount == 1)
   {
      sdChanged = 2; // Compensate for the extra OnCalculate call during OnInit
   }
#endif
   if (sdChanged > 0 || time0 != stPrevTime0)
   {
      if (sdChanged > 0 && LogLevel >= LogDebug) { Print("OnTick TickCount=", stTickCount, " sdChanged=", sdChanged); }

      // For optimal performance - Comment out variables that are not required
      bool    newBar = time0 != stPrevTime0;
      int     overallTrend = (int)iCustomSd(SD_OVERALLTREND);  // Overall Trend
      int     trend1       = (int)iCustomSd(SD_TREND1);        // Timeframe 1
      int     engulfing1   = (int)iCustomSd(SD_ENGULFING1);
      double  qTop1        = iCustomSd(SD_QTOP1);
      double  qBottom1     = iCustomSd(SD_QBOTTOM1);
      int     supDemHint1  = (int)iCustomSd(SD_SUPDEMHINT1);
      int     trend2       = (int)iCustomSd(SD_TREND2);        // Timeframe 2
      int     engulfing2   = (int)iCustomSd(SD_ENGULFING2);
      double  qTop2        = iCustomSd(SD_QTOP2);
      double  qBottom2     = iCustomSd(SD_QBOTTOM2);
      int     supDemHint2  = (int)iCustomSd(SD_SUPDEMHINT2);
      int     trend3       = (int)iCustomSd(SD_TREND3);        // Timeframe 3
      int     engulfing3   = (int)iCustomSd(SD_ENGULFING3);
      double  qTop3        = iCustomSd(SD_QTOP3);
      double  qBottom3     = iCustomSd(SD_QBOTTOM3);
      
      // Calculate the QHi and QLow values
      double  qHi1         = qTop1 * 0.75 + qBottom1 * 0.25;
      double  qHi2         = qTop2 * 0.75 + qBottom2 * 0.25;
      double  qHi3         = qTop3 * 0.75 + qBottom3 * 0.25;
      double  qLow1        = qTop1 * 0.25 + qBottom1 * 0.75;
      double  qLow2        = qTop2 * 0.25 + qBottom2 * 0.75;
      double  qLow3        = qTop3 * 0.25 + qBottom3 * 0.75;
      
      if (LogLevel >= LogAll) { Print("OnTick sdChanged=", sdChanged); }
      
      static datetime stPrevTime2_0;
      static datetime stPrevTime3_0;
      
      if (sdChanged > 1)
      {
         PopulateLevels(ExtSupLevels1, ExtDemLevels1, ExtLevelIndex1);
         PopulateLevels(ExtSupLevels2, ExtDemLevels2, ExtLevelIndex2);
         PopulateLevels(ExtSupLevels3, ExtDemLevels3, ExtLevelIndex3);
         stPrevTime2_0 = ExtLevelIndex2.IsValid() ? iTime(NULL, ExtLevelIndex2.Timeframe, 0) : 0;
         stPrevTime3_0 = ExtLevelIndex3.IsValid() ? iTime(NULL, ExtLevelIndex3.Timeframe, 0) : 0;
      }
      else
      {
         bool updateAll = UpdateRetouches == SD_EA_RETOUCH_ALL;
         
         if (newBar)
         {
            IncrementalLevels(ExtSupLevels1, ExtDemLevels1, ExtLevelIndex1, updateAll, newBar); // Populate incremental levels
            
            if (ValidateActive && ! ValidateLevels(ExtSupLevels1, ExtDemLevels1, ExtLevelIndex1, "1", updateAll))
            {
               PopulateLevels(ExtSupLevels1, ExtDemLevels1, ExtLevelIndex1);
               if (ValidateIncremental == SD_EA_VAL_FIRST) { ValidateActive = false; }
            }
            // Timeframe 2
            if (ExtLevelIndex2.IsValid() && ExtLevelIndex2.Timeframe != ExtLevelIndex1.Timeframe)
            {
               datetime time2_0 = iTime(NULL, ExtLevelIndex2.Timeframe, 0);
               bool     newBar2 = stPrevTime2_0 != 0 && stPrevTime2_0 != time2_0;
               
               //if (newBar2 && time2_0 > D'2016.12.11 22:54' && time2_0 <  D'2016.12.11 23:10') { Print("newBar2 stPrevTime2_0=", stPrevTime2_0, "  time2_0=", time2_0); }
               IncrementalLevels(ExtSupLevels2, ExtDemLevels2, ExtLevelIndex2, updateAll, newBar2); // Populate incremental levels
               
               if (ValidateActive && !ValidateLevels(ExtSupLevels2, ExtDemLevels2, ExtLevelIndex2, "2", updateAll))
               {
                  PopulateLevels(ExtSupLevels2, ExtDemLevels2, ExtLevelIndex2);
                  if (ValidateIncremental == SD_EA_VAL_FIRST) { ValidateActive = false; }
               }
               if (stPrevTime2_0 != time2_0)
               {
                  stPrevTime2_0 = time2_0;
               }
            }
            // Timeframe 3
            if (ExtLevelIndex3.IsValid() && ExtLevelIndex3.Timeframe != ExtLevelIndex1.Timeframe)
            {
               datetime time3_0 = iTime(NULL, ExtLevelIndex3.Timeframe, 0);
               bool     newBar3 = stPrevTime3_0 != 0 && stPrevTime3_0 != time3_0;
               
               IncrementalLevels(ExtSupLevels3, ExtDemLevels3, ExtLevelIndex3, updateAll, newBar3); // Populate incremental levels
               
               if (ValidateActive && !ValidateLevels(ExtSupLevels3, ExtDemLevels3, ExtLevelIndex3, "3", updateAll))
               {
                  PopulateLevels(ExtSupLevels3, ExtDemLevels3, ExtLevelIndex3);
                  if (ValidateIncremental == SD_EA_VAL_FIRST) { ValidateActive = false; }
               }
               if (stPrevTime3_0 != time3_0)
               {
                  stPrevTime3_0 = time3_0;
               }
            }
         }
         else
         {
            IncrementalLevels(ExtSupLevels1, ExtDemLevels1, ExtLevelIndex1);  // Populate incremental levels
            IncrementalLevels(ExtSupLevels2, ExtDemLevels2, ExtLevelIndex2);  // Populate incremental levels
            IncrementalLevels(ExtSupLevels3, ExtDemLevels3, ExtLevelIndex3);  // Populate incremental levels
         }
      }
      if (newBar)
      {
         // Add user code for new bars
         stPrevTime0 = time0;  // Update prevTime0
      }
      // Add user code for each tick when SdChanged
      //
      //if (ArraySize(ExtSupLevels1) > 1)
      //{
      //   SupDemLevel* secondSupLevel = ExtSupLevels1[1];
      //   double       secondSupLevelHigh = secondSupLevel.PriceHigh;
      //   
      //}
   }
   // Add user code for each tick
}


void PopulateLevels(SupDemLevel* &supLevels[], SupDemLevel* &demLevels[], const SupDemLevelIdx& levelIdx)
{
   if (levelIdx.IsValid())
   {
      ArrayDelete(supLevels);
      ArrayDelete(demLevels);
      PopulateLevelsSupplies(supLevels, levelIdx);
      PopulateLevelsDemands(demLevels, levelIdx);
      if (LogLevel >= LogInfo) { Print(TimeframeToString(levelIdx.Timeframe), " PopulateLevels NumSups=", ArraySize(supLevels), " NumDems=", ArraySize(demLevels)); }
   }
}


void PopulateLevelsSupplies(SupDemLevel* &levels[], const SupDemLevelIdx& levelIdx)
{
   int i = 0;
   int barNumOffset = (int)iCustomSd(levelIdx.SupBarNum, i);
   
   while (barNumOffset > 0 && barNumOffset != EMPTY_VALUE)
   {
      i += barNumOffset;
      ArrayAppend(levels, CreateNewSupply(levelIdx, i));
      barNumOffset = (int)iCustomSd(levelIdx.SupBarNum, i);
   }
   if (LogLevel >= LogDebug) { Print(TimeframeToString(levelIdx.Timeframe), " PopulateLevelsSupplies NumSups=", ArraySize(levels)); }
}


void PopulateLevelsDemands(SupDemLevel* &levels[], const SupDemLevelIdx& levelIdx)
{
   int i = 0;
   int barNumOffset = (int)iCustomSd(levelIdx.DemBarNum, i);
   
   while (barNumOffset > 0 && barNumOffset != EMPTY_VALUE)
   {
      i += barNumOffset;
      ArrayAppend(levels, CreateNewDemand(levelIdx, i));
      barNumOffset = (int)iCustomSd(levelIdx.DemBarNum, i);
   }
   if (LogLevel >= LogDebug) { Print(TimeframeToString(levelIdx.Timeframe), " PopulateLevelsDemands NumDems=", ArraySize(levels)); }
}


void IncrementalLevels(SupDemLevel* &supLevels[], SupDemLevel* &demLevels[], const SupDemLevelIdx& levelIdx, bool updateRetouchesAll = false, bool isNewBar = false)
{
   if (levelIdx.IsValid())
   {
      if (isNewBar)
      {
         int numSups = ArraySize(supLevels);
         int numDems = ArraySize(demLevels);
         
         for(int i = 0; i < numSups; i++)
         {
            supLevels[i].IncrementBarNums();
         }
         for(int i = 0; i < numDems; i++)
         {
            demLevels[i].IncrementBarNums();;
         }
      }
      IncrementalSupplies(supLevels, levelIdx, updateRetouchesAll);
      IncrementalDemands(demLevels, levelIdx, updateRetouchesAll);
   }
}


void IncrementalSupplies(SupDemLevel* &supLevels[], const SupDemLevelIdx& levelIdx, bool updateRetouchesAll)
{
   int numSups = ArraySize(supLevels);
   int bufferSupBarNum = (int)iCustomSd(levelIdx.SupBarNum, 0);
   
   if (bufferSupBarNum > 0 && bufferSupBarNum != EMPTY_VALUE)
   {
      if (numSups < 1)
      {
         PopulateLevelsSupplies(supLevels, levelIdx);
      }
      else
      {
         int newBarNum = 0;
         int supBarNum = supLevels[0].BarNum;
         
         if (bufferSupBarNum < supBarNum)                                     // New level detected
         {
            newBarNum = bufferSupBarNum;                                      // Set the newBarNum
            int barNumOffset = (int)iCustomSd(levelIdx.SupBarNum, bufferSupBarNum); // Get the next BarNum offset from the buffer
            
            if (barNumOffset > 0 && barNumOffset != EMPTY_VALUE)
            {
               bufferSupBarNum += barNumOffset;
               if (LogLevel >= LogInfo && bufferSupBarNum > supBarNum) { Print(TimeframeToString(levelIdx.Timeframe), " Adding new Sup BarNum=", newBarNum, " and removing old Sup BarNum=", supBarNum); }
            }
            else
            {
               bufferSupBarNum = -1;
               ArrayDelete(supLevels);
            }
         }
         if (bufferSupBarNum > supBarNum)
         {
            int numToDelete = 1;
            
            while (numToDelete < numSups && bufferSupBarNum > supLevels[numToDelete].BarNum)
            {
               numToDelete++;
            }
            ArrayDeleteFromStart(supLevels, numToDelete);
            if (LogLevel >= LogDebug) { Print(TimeframeToString(levelIdx.Timeframe), " Sup-", numToDelete, " => ", ArraySize(supLevels)); }
         }
         if (newBarNum > 0)
         {
            ArrayPrepend(supLevels, CreateNewSupply(levelIdx, newBarNum));
            if (LogLevel >= LogDebug) { Print(TimeframeToString(levelIdx.Timeframe), " Sup+1 => ", ArraySize(supLevels)); }
         }
         else if (levelIdx.PopulateScoresAndRetouches() && ArraySize(supLevels) > 0)
         {
            SupDemLevel* firstSup = supLevels[0];
            
            firstSup.Score = (int)iCustomSd(SD_SUPSCORE, 0);
            firstSup.Retouch = (int)iCustomSd(SD_SUPRETOUCH, 0);
            firstSup.IsStrong = (bool)iCustomSd(SD_SUPSTRONG, 0);
         }
         if (updateRetouchesAll && levelIdx.PopulateScoresAndRetouches())
         {
            numSups = ArraySize(supLevels);
            for(int i = 1; i < numSups; i++)
            {
               UpdateRetouchesSupply(supLevels[i]);
            }
         }
      }
   }
   else if (numSups > 0)
   {
      ArrayDelete(supLevels);
   }
}


void IncrementalDemands(SupDemLevel* &demLevels[], const SupDemLevelIdx& levelIdx, bool updateRetouchesAll)
{
   int numDems = ArraySize(demLevels);
   int bufferDemBarNum = (int)iCustomSd(levelIdx.DemBarNum, 0);
   
   if (bufferDemBarNum > 0 && bufferDemBarNum != EMPTY_VALUE)
   {
      if (numDems < 1)
      {
         PopulateLevelsDemands(demLevels, levelIdx);
      }
      else
      {
         int newBarNum = 0;
         int demBarNum = demLevels[0].BarNum;
         
         if (bufferDemBarNum < demBarNum)                                     // New level detected
         {
            newBarNum = bufferDemBarNum;                                      // Set the newBarNum
            int barNumOffset = (int)iCustomSd(levelIdx.DemBarNum, bufferDemBarNum); // Get the next BarNum offset from the buffer
            
            if (barNumOffset > 0 && barNumOffset != EMPTY_VALUE)
            {
               bufferDemBarNum += barNumOffset;
               if (LogLevel >= LogInfo && bufferDemBarNum > demBarNum) { Print(TimeframeToString(levelIdx.Timeframe), " Adding new Dem BarNum=", newBarNum, " and removing old Dem BarNum=", demBarNum); }
            }
            else
            {
               bufferDemBarNum = -1;
               ArrayDelete(demLevels);
            }
         }
         if (bufferDemBarNum > demBarNum)
         {
            int numToDelete = 1;
            
            while (numToDelete < numDems && bufferDemBarNum > demLevels[numToDelete].BarNum)
            {
               numToDelete++;
            }
            ArrayDeleteFromStart(demLevels, numToDelete);
            if (LogLevel >= LogDebug) { Print(TimeframeToString(levelIdx.Timeframe), " Dem-", numToDelete, " => ", ArraySize(demLevels)); }
         }
         if (newBarNum > 0)
         {
            ArrayPrepend(demLevels, CreateNewDemand(levelIdx, newBarNum));
            if (LogLevel >= LogDebug) { Print(TimeframeToString(levelIdx.Timeframe), " Dem+1 => ", ArraySize(demLevels)); }
         }
         else if (levelIdx.PopulateScoresAndRetouches() && ArraySize(demLevels) > 0)
         {
            SupDemLevel* firstDem = demLevels[0];
            
            firstDem.Score = (int)iCustomSd(SD_DEMSCORE, 0);
            firstDem.Retouch = (int)iCustomSd(SD_DEMRETOUCH, 0);
            firstDem.IsStrong = (bool)iCustomSd(SD_DEMSTRONG, 0);
         }
         if (updateRetouchesAll && levelIdx.PopulateScoresAndRetouches())
         {
            numDems = ArraySize(demLevels);
            for(int i = 1; i < numDems; i++)
            {
               UpdateRetouchesDemand(demLevels[i]);
            }
         }
      }
   }
   else if (numDems > 0)
   {
      ArrayDelete(demLevels);
   }
}


bool ValidateLevels(SupDemLevel* &supLevels[], SupDemLevel* &demLevels[], const SupDemLevelIdx& levelIdx, string prefix, bool validateAll = false)
{
   bool         ret = true;
   SupDemLevel* validateSups[];
   SupDemLevel* validateDems[];
   
   PopulateLevelsSupplies(validateSups, levelIdx);
   PopulateLevelsDemands(validateDems, levelIdx);
   
   if (!SupDemLevel::CompareLevels(supLevels, validateSups, !validateAll, LogLevel))
   {
      if (LogLevel >= LogError) { Print(TimeframeToString(levelIdx.Timeframe), " ", prefix, " ERROR Sup Levels differ time0=", iTime(NULL, levelIdx.Timeframe, 0)); }
      ret = false;
   }
   if (!SupDemLevel::CompareLevels(demLevels, validateDems, !validateAll, LogLevel))
   {
      if (LogLevel >= LogError) { Print(TimeframeToString(levelIdx.Timeframe), " ", prefix, " ERROR Dem Levels differ time0=", iTime(NULL, levelIdx.Timeframe, 0)); }
      ret = false;
   }
   ArrayDelete(validateSups);
   ArrayDelete(validateDems);
   return ret;
}


SupDemLevel* CreateNewSupply(const SupDemLevelIdx& levelIdx, int barNum)
{
   SupDemLevel* ret = new SupDemLevel();
   
   ret.BarNum = barNum;
   ret.Timestamp = iTime(NULL, levelIdx.Timeframe, barNum);
   ret.PriceHigh = iHigh(NULL, levelIdx.Timeframe, barNum);
   ret.PriceLow = iCustomSd(levelIdx.SupLow, barNum);
   
   if (levelIdx.PopulateScoresAndRetouches())
   {
      ret.Score = (int)iCustomSd(SD_SUPSCORE, barNum);
      ret.IsStrong = (bool)iCustomSd(SD_SUPSTRONG, barNum);
      ret.Retouch = (int)iCustomSd(SD_SUPRETOUCH, barNum);
   }
   return ret;
}


SupDemLevel* CreateNewDemand(const SupDemLevelIdx& levelIdx, int barNum)
{
   SupDemLevel* ret = new SupDemLevel();
   
   ret.BarNum = barNum;
   ret.Timestamp = iTime(NULL, levelIdx.Timeframe, barNum);
   ret.PriceHigh = iCustomSd(levelIdx.DemHigh, barNum);
   ret.PriceLow = iLow(NULL, levelIdx.Timeframe, barNum);
   
   if (levelIdx.PopulateScoresAndRetouches())
   {
      ret.Score = (int)iCustomSd(SD_DEMSCORE, barNum);
      ret.IsStrong = (bool)iCustomSd(SD_DEMSTRONG, barNum);
      ret.Retouch = (int)iCustomSd(SD_DEMRETOUCH, barNum);
   }
   return ret;
}


void UpdateRetouchesSupply(SupDemLevel& sup)
{
   int barNum = sup.BarNum;
   
   sup.Score = (int)iCustomSd(SD_SUPSCORE, barNum);
   sup.IsStrong = (bool)iCustomSd(SD_SUPSTRONG, barNum);
   sup.Retouch = (int)iCustomSd(SD_SUPRETOUCH, barNum);
}


void UpdateRetouchesDemand(SupDemLevel& dem)
{
   int barNum = dem.BarNum;
   
   dem.Score = (int)iCustomSd(SD_DEMSCORE, barNum);
   dem.IsStrong = (bool)iCustomSd(SD_DEMSTRONG, barNum);
   dem.Retouch = (int)iCustomSd(SD_DEMRETOUCH, barNum);
}
