
#property description "Blahtech Supply Demand - iCustom Indicator Example"
#property version     "2.5"
#property strict

#property indicator_chart_window

#include "BlahtechUtils.mqh"

#ifdef __MQL4__
input string          IndicatorPath = "Market\\Blahtech Supply Demand";          // Indicator Path
#else
input string          IndicatorPath = "Market\\Blahtech Supply Demand MT5.ex5";  // Indicator Path
#endif
input SdEaModeEnum  ExpertMode = SD_EA_EXPERT_NO_DRAW;                           // Expert Mode
input int           MinBarsLevel = 3;                                            // Min Bars For Level
input int           MaxBars = 20000;                                             // Max Bars
input int           ForceRecalcBars = 10000;                                     // Force Recalc Bars
input string        LoadParameterFile = "SD_EA_Params.set";                      // Parameter File/Values
//input string            LoadParameterFile =  "values:DetectionMethod2=0;Hud2=2";   // Parameter File/Values
input LogLevelEnum  LogLevel = LogError;                                         // Log Level

int    ExtIndicatorHandle;


double iCustomSd(SdBufferIndexEnum index, int shift = 0)
{
#ifdef __MQL4__
   return iCustom(NULL, PERIOD_CURRENT, IndicatorPath, ExpertMode, MinBarsLevel, MaxBars, ForceRecalcBars, LoadParameterFile, index, shift);
#else
   double arr[1];
   return CopyBuffer(ExtIndicatorHandle, index, shift, 1, arr) > 0 ? arr[0] : 0.0;
#endif
}



int OnInit()
{
#ifdef __MQL4__
   int  ret = INIT_SUCCEEDED;
#else
   ExtIndicatorHandle = iCustom(NULL, PERIOD_CURRENT, IndicatorPath, ExpertMode, MinBarsLevel, MaxBars, ForceRecalcBars, LoadParameterFile);
   int  ret = ExtIndicatorHandle > 0 ? INIT_SUCCEEDED : INIT_FAILED;
#endif
   if (LogLevel >= LogInfo) { Print("OnInit - IndicatorPath=", IndicatorPath); }
   return ret;
}


void OnDeinit(const int reason)
{
   #ifndef __MQL4__
   if (ExtIndicatorHandle > 0)
   {
      IndicatorRelease(ExtIndicatorHandle);
   }
   #endif
}


int OnCalculate(
   const int rates_total,
   const int prev_calculated,
   const datetime &time[],
   const double &open[],
   const double &high[],
   const double &low[],
   const double &close[],
   const long &tick_volume[],
   const long &volume[],
   const int &spread[])
{
   static int      stTickCount;
   static datetime stPrevTime0;
   #ifdef __MQL4__
   datetime        time0 = time[0];
   #else
   datetime        time0 = time[rates_total-1];
   #endif
   int             sdChanged = (int)iCustomSd(SD_CHANGED);  // Only do work when values have changed

   stTickCount++;
#ifndef __MQL4__
   if (stTickCount == 1)
   {
      sdChanged = 2; // Compensate for the extra OnCalculate call during OnInit
   }
#endif   
   if (sdChanged > 0 || time0 != stPrevTime0)
   {
      if (sdChanged > 0 && LogLevel >= LogDebug) { Print("OnCalculate TickCount=", stTickCount, " sdChanged=", sdChanged); }

      // For optimal performance - Comment out variables that are not required
      bool     newBar = time0 != stPrevTime0;
      int      supBarNum    = (int)iCustomSd(SD_SUPBARNUM);       // First Supply
      bool     hasSup       = supBarNum > 0 && supBarNum != EMPTY_VALUE;
      datetime supTimestamp = hasSup ? iTime(NULL, PERIOD_CURRENT, supBarNum) : 0;
      double   supPriceHigh = hasSup ? iHigh(NULL, PERIOD_CURRENT, supBarNum) : 0;
      double   supPriceLow  = hasSup ? iCustomSd(SD_SUPLOW) : 0;
      int      supScore     = hasSup ? (int)iCustomSd(SD_SUPSCORE) : 0;
      int      supIsStrong  = hasSup ? (bool)iCustomSd(SD_SUPSTRONG) : false;
      int      supRetouch   = hasSup ? (int)iCustomSd(SD_SUPRETOUCH) : 0;
      
      int      demBarNum    = (int)iCustomSd(SD_DEMBARNUM);       // First Demand
      bool     hasDem       = demBarNum > 0 && demBarNum != EMPTY_VALUE;
      datetime demTimestamp = hasDem ? iTime(NULL, PERIOD_CURRENT, demBarNum) : 0;
      double   demPriceHigh = hasDem ? iCustomSd(SD_DEMHIGH) : 0;
      double   demPriceLow  = hasDem ? iLow(NULL, PERIOD_CURRENT, demBarNum) : 0;
      int      demScore     = hasDem ? (int)iCustomSd(SD_DEMSCORE) : 0;
      int      demIsStrong  = hasDem ? (bool)iCustomSd(SD_DEMSTRONG) : false;
      int      demRetouch   = hasDem ? (int)iCustomSd(SD_DEMRETOUCH) : 0;
      
      int      overallTrend = (int)iCustomSd(SD_OVERALLTREND);    // Overall Trend
      int      trend1       = (int)iCustomSd(SD_TREND1);          // Timeframe 1
      int      engulfing1   = (int)iCustomSd(SD_ENGULFING1);
      double   qTop1        = iCustomSd(SD_QTOP1);
      double   qBottom1     = iCustomSd(SD_QBOTTOM1);
      int      supDemHint1  = (int)iCustomSd(SD_SUPDEMHINT1);
      int      trend2       = (int)iCustomSd(SD_TREND2);          // Timeframe 2
      int      engulfing2   = (int)iCustomSd(SD_ENGULFING2);
      double   qTop2        = iCustomSd(SD_QTOP2);
      double   qBottom2     = iCustomSd(SD_QBOTTOM2);
      int      supDemHint2  = (int)iCustomSd(SD_SUPDEMHINT2);
      int      trend3       = (int)iCustomSd(SD_TREND3);          // Timeframe 3
      int      engulfing3   = (int)iCustomSd(SD_ENGULFING3);
      double   qTop3        = iCustomSd(SD_QTOP3);
      double   sdBottom3    = iCustomSd(SD_QBOTTOM3);

      if (LogLevel >= LogAll) { Print("OnCalculate sdChanged=", sdChanged); }
      
      if (newBar)
      {
         // Add user code for new bars
         stPrevTime0 = time0;  // Update prevTime0
      }
      // Add user code for each tick when SdChanged
   }
   // Add user code for each tick
   return rates_total;
}
